@@ -1,5 +1,5 @@
 #include "moar.h"
-#include "math.h"
+#include <math.h>
 
 /* This representation's function pointer table. */
 static const MVMREPROps this_repr;
@@ -114,6 +114,7 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
     /* If we have no attributes in the index mapping, then just the header. */
     if (repr_data->name_to_index_mapping[0].class_key == NULL) {
         repr_data->struct_size = 1; /* avoid 0-byte malloc */
+        repr_data->struct_align = ALIGNOF(void *);
     }
 
     /* Otherwise, we need to compute the allocation strategy.  */
@@ -122,7 +123,7 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
         MVMint32 cur_size    = 0;
         /* The structure itself will be the multiple of its biggest element in size.
          * So we keep track of that biggest element. */
-        MVMint32 multiple_of = 0;
+        MVMint32 multiple_of = 1;
 
         /* Get number of attributes and set up various counters. */
         MVMint32 num_attrs        = MVM_repr_elems(tc, flat_list);
@@ -137,6 +138,7 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
         repr_data->struct_offsets      = (MVMint32 *)   MVM_malloc(info_alloc * sizeof(MVMint32));
         repr_data->flattened_stables   = (MVMSTable **) MVM_calloc(info_alloc, sizeof(MVMObject *));
         repr_data->member_types        = (MVMObject **) MVM_calloc(info_alloc, sizeof(MVMObject *));
+        repr_data->struct_align        = ALIGNOF(void *);
 
         /* Go over the attributes and arrange their allocation. */
         for (i = 0; i < num_attrs; i++) {
@@ -198,10 +200,22 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
                     if (inlined) {
                         MVMCStructREPRData *cstruct_repr_data = (MVMCStructREPRData *)STABLE(type)->REPR_data;
                         bits                                  = cstruct_repr_data->struct_size * 8;
-                        align                                 = cstruct_repr_data->struct_size;
+                        align                                 = cstruct_repr_data->struct_align;
                         repr_data->attribute_locations[i]    |= MVM_CSTRUCT_ATTR_INLINED;
                     }
                 }
+                else if (type_id == MVM_REPR_ID_MVMCPPStruct) {
+                    /* It's a CPPStruct. */
+                    repr_data->num_child_objs++;
+                    repr_data->attribute_locations[i] = (cur_obj_attr++ << MVM_CSTRUCT_ATTR_SHIFT) | MVM_CSTRUCT_ATTR_CPPSTRUCT;
+                    repr_data->member_types[i] = type;
+                    if (inlined) {
+                        MVMCPPStructREPRData *cppstruct_repr_data = (MVMCPPStructREPRData *)STABLE(type)->REPR_data;
+                        bits                                      = cppstruct_repr_data->struct_size * 8;
+                        align                                     = cppstruct_repr_data->struct_align;
+                        repr_data->attribute_locations[i]        |= MVM_CSTRUCT_ATTR_INLINED;
+                    }
+                }
                 else if (type_id == MVM_REPR_ID_MVMCUnion) {
                     /* It's a CUnion. */
                     repr_data->num_child_objs++;
@@ -210,7 +224,7 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
                     if (inlined) {
                         MVMCUnionREPRData *cunion_repr_data = (MVMCUnionREPRData *)STABLE(type)->REPR_data;
                         bits                                = cunion_repr_data->struct_size * 8;
-                        align                               = cunion_repr_data->struct_size;
+                        align                               = cunion_repr_data->struct_align;
                         repr_data->attribute_locations[i]  |= MVM_CSTRUCT_ATTR_INLINED;
                     }
                 }
@@ -222,7 +236,7 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
                 }
                 else {
                     MVM_exception_throw_adhoc(tc,
-                        "CStruct representation only handles int, num, CArray, CPointer and CStruct");
+                        "CStruct representation only handles int, num, CArray, CPointer, CStruct, CPPStruct and CUnion");
                 }
             }
             else {
@@ -243,6 +257,9 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
                 cur_size += align - cur_size % align;
             }
 
+            if (align > repr_data->struct_align)
+                repr_data->struct_align = align;
+
             repr_data->struct_offsets[i] = cur_size;
             cur_size += bits / 8;
 
@@ -344,8 +361,10 @@ static void copy_to(MVMThreadContext *tc, MVMSTable *st, void *src, MVMObject *d
 /* Helper for complaining about attribute access errors. */
 MVM_NO_RETURN static void no_such_attribute(MVMThreadContext *tc, const char *action, MVMObject *class_handle, MVMString *name) MVM_NO_RETURN_GCC;
 static void no_such_attribute(MVMThreadContext *tc, const char *action, MVMObject *class_handle, MVMString *name) {
-    MVM_exception_throw_adhoc(tc, "Can not %s non-existent attribute '%s'",
-        action, MVM_string_utf8_encode_C_string(tc, name));
+    char *c_name = MVM_string_utf8_encode_C_string(tc, name);
+    char *waste[] = { c_name, NULL };
+    MVM_exception_throw_adhoc_free(tc, waste, "Can not %s non-existent attribute '%s'",
+        action, c_name);
 }
 
 /* Helper to die because this type doesn't support attributes. */
@@ -363,7 +382,7 @@ static void get_attribute(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
     MVMint64 slot;
 
     if (!repr_data)
-        MVM_exception_throw_adhoc(tc, "P6opaque: must compose before using get_attribute");
+        MVM_exception_throw_adhoc(tc, "CStruct: must compose before using get_attribute");
 
     slot = hint >= 0 ? hint : try_get_slot(tc, repr_data, class_handle, name);
     if (slot >= 0) {
@@ -395,6 +414,13 @@ static void get_attribute(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
                             else
                                 obj = MVM_nativecall_make_cstruct(tc, typeobj, cobj);
                         }
+                        else if(type == MVM_CSTRUCT_ATTR_CPPSTRUCT) {
+                            if (repr_data->attribute_locations[slot] & MVM_CSTRUCT_ATTR_INLINED)
+                                obj = MVM_nativecall_make_cppstruct(tc, typeobj,
+                                    (char *)body->cstruct + repr_data->struct_offsets[slot]);
+                            else
+                                obj = MVM_nativecall_make_cppstruct(tc, typeobj, cobj);
+                        }
                         else if(type == MVM_CSTRUCT_ATTR_CUNION) {
                             if (repr_data->attribute_locations[slot] & MVM_CSTRUCT_ATTR_INLINED)
                                 obj = MVM_nativecall_make_cunion(tc, typeobj,
@@ -466,7 +492,7 @@ static void bind_attribute(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
     MVMint64 slot;
 
     if (!repr_data)
-        MVM_exception_throw_adhoc(tc, "P6opaque: must compose before using bind_attribute");
+        MVM_exception_throw_adhoc(tc, "CStruct: must compose before using bind_attribute");
 
     slot = hint >= 0 ? hint : try_get_slot(tc, repr_data, class_handle, name);
     if (slot >= 0) {
@@ -501,6 +527,12 @@ static void bind_attribute(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
                                 "Can only store CStruct attribute in CStruct slot in CStruct");
                         cobj = ((MVMCStruct *)value)->body.cstruct;
                     }
+                    else if (type == MVM_CSTRUCT_ATTR_CPPSTRUCT) {
+                        if (REPR(value)->ID != MVM_REPR_ID_MVMCPPStruct)
+                            MVM_exception_throw_adhoc(tc,
+                                "Can only store CPPStruct attribute in CPPStruct slot in CStruct");
+                        cobj = ((MVMCPPStruct *)value)->body.cppstruct;
+                    }
                     else if (type == MVM_CSTRUCT_ATTR_CUNION) {
                         if (REPR(value)->ID != MVM_REPR_ID_MVMCUnion)
                             MVM_exception_throw_adhoc(tc,
@@ -607,6 +639,26 @@ static void gc_mark_repr_data(MVMThreadContext *tc, MVMSTable *st, MVMGCWorklist
     }
 }
 
+/* Free representation data. */
+static void gc_free_repr_data(MVMThreadContext *tc, MVMSTable *st) {
+    MVMCStructREPRData *repr_data = (MVMCStructREPRData *)st->REPR_data;
+
+    /* May not have survived to composition. */
+    if (repr_data == NULL)
+        return;
+
+    if (repr_data->name_to_index_mapping) {
+        MVM_free(repr_data->name_to_index_mapping);
+        MVM_free(repr_data->attribute_locations);
+        MVM_free(repr_data->struct_offsets);
+        MVM_free(repr_data->flattened_stables);
+        MVM_free(repr_data->member_types);
+        MVM_free(repr_data->initialize_slots);
+    }
+
+    MVM_free(st->REPR_data);
+}
+
 /* This is called to do any cleanup of resources when an object gets
  * embedded inside another one. Never called on a top-level object. */
 static void gc_cleanup(MVMThreadContext *tc, MVMSTable *st, void *data) {
@@ -696,7 +748,7 @@ static void deserialize_repr_data(MVMThreadContext *tc, MVMSTable *st, MVMSerial
         repr_data->struct_offsets[i] = MVM_serialization_read_varint(tc, reader);
 
         if(MVM_serialization_read_varint(tc, reader)){
-            repr_data->flattened_stables[i] = MVM_serialization_read_stable_ref(tc, reader);
+            MVM_ASSIGN_REF(tc, &(st->header), repr_data->flattened_stables[i], MVM_serialization_read_stable_ref(tc, reader));
         }
         else {
             repr_data->flattened_stables[i] = NULL;
@@ -759,7 +811,7 @@ static const MVMREPROps this_repr = {
     gc_free,
     gc_cleanup,
     gc_mark_repr_data,
-    NULL, /* gc_free_repr_data */
+    gc_free_repr_data,
     compose,
     NULL, /* spesh */
     "CStruct", /* name */
