@@ -1,5 +1,4 @@
 #include "moar.h"
-#include "math.h"
 
 /* This representation's function pointer table. */
 static const MVMREPROps this_repr;
@@ -114,6 +113,7 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
     /* If we have no attributes in the index mapping, then just the header. */
     if (repr_data->name_to_index_mapping[0].class_key == NULL) {
         repr_data->struct_size = 1; /* avoid 0-byte malloc */
+        repr_data->struct_align = ALIGNOF(void *);
     }
 
     /* Otherwise, we need to compute the allocation strategy.  */
@@ -197,9 +197,22 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
                     if (inlined) {
                         MVMCStructREPRData *cstruct_repr_data = (MVMCStructREPRData *)STABLE(type)->REPR_data;
                         bits                                  = cstruct_repr_data->struct_size * 8;
+                        align                                 = cstruct_repr_data->struct_align;
                         repr_data->attribute_locations[i]    |= MVM_CUNION_ATTR_INLINED;
                     }
                 }
+                else if (type_id == MVM_REPR_ID_MVMCPPStruct) {
+                    /* It's a CPPStruct. */
+                    repr_data->num_child_objs++;
+                    repr_data->attribute_locations[i] = (cur_obj_attr++ << MVM_CUNION_ATTR_SHIFT) | MVM_CUNION_ATTR_CPPSTRUCT;
+                    repr_data->member_types[i] = type;
+                    if (inlined) {
+                        MVMCPPStructREPRData *cppstruct_repr_data = (MVMCPPStructREPRData *)STABLE(type)->REPR_data;
+                        bits                                      = cppstruct_repr_data->struct_size * 8;
+                        align                                     = cppstruct_repr_data->struct_align;
+                        repr_data->attribute_locations[i]        |= MVM_CUNION_ATTR_INLINED;
+                    }
+                }
                 else if (type_id == MVM_REPR_ID_MVMCUnion) {
                     /* It's a CUnion. */
                     repr_data->num_child_objs++;
@@ -208,6 +221,7 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
                     if (inlined) {
                         MVMCUnionREPRData *cunion_repr_data = (MVMCUnionREPRData *)STABLE(type)->REPR_data;
                         bits                                = cunion_repr_data->struct_size * 8;
+                        align                               = cunion_repr_data->struct_align;
                         repr_data->attribute_locations[i]  |= MVM_CUNION_ATTR_INLINED;
                     }
                 }
@@ -219,7 +233,7 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
                 }
                 else {
                     MVM_exception_throw_adhoc(tc,
-                        "CUnion representation only handles int, num, CArray, CPointer and CStruct");
+                        "CUnion representation only handles int, num, CArray, CPointer, CStruct, CPPStruct and CUnion");
                 }
             }
             else {
@@ -232,6 +246,9 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
                     "CUnion only supports native types that are a multiple of 8 bits wide (was passed: %"PRId32")", bits);
             }
 
+            if (align > repr_data->struct_align)
+                repr_data->struct_align = align;
+
             repr_data->struct_offsets[i] = 0;
             if (bits / 8 > total_size)
                 total_size = bits / 8;
@@ -357,8 +374,10 @@ static void copy_to(MVMThreadContext *tc, MVMSTable *st, void *src, MVMObject *d
 /* Helper for complaining about attribute access errors. */
 MVM_NO_RETURN static void no_such_attribute(MVMThreadContext *tc, const char *action, MVMObject *class_handle, MVMString *name) MVM_NO_RETURN_GCC;
 static void no_such_attribute(MVMThreadContext *tc, const char *action, MVMObject *class_handle, MVMString *name) {
-    MVM_exception_throw_adhoc(tc, "Can not %s non-existent attribute '%s'",
-        action, MVM_string_utf8_encode_C_string(tc, name));
+    char *c_name = MVM_string_utf8_encode_C_string(tc, name);
+    char *waste[] = { c_name, NULL };
+    MVM_exception_throw_adhoc_free(tc, waste, "Can not %s non-existent attribute '%s'",
+        action, c_name);
 }
 
 /* Helper to die because this type doesn't support attributes. */
@@ -376,7 +395,7 @@ static void get_attribute(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
     MVMint64 slot;
 
     if (!repr_data)
-        MVM_exception_throw_adhoc(tc, "P6opaque: must compose before using get_attribute");
+        MVM_exception_throw_adhoc(tc, "CUnion: must compose before using get_attribute");
 
     slot = hint >= 0 ? hint : try_get_slot(tc, repr_data, class_handle, name);
     if (slot >= 0) {
@@ -408,6 +427,13 @@ static void get_attribute(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
                             else
                                 obj = MVM_nativecall_make_cstruct(tc, typeobj, cobj);
                         }
+                        else if(type == MVM_CUNION_ATTR_CPPSTRUCT) {
+                            if (repr_data->attribute_locations[slot] & MVM_CUNION_ATTR_INLINED)
+                                obj = MVM_nativecall_make_cppstruct(tc, typeobj,
+                                    (char *)body->cunion + repr_data->struct_offsets[slot]);
+                            else
+                                obj = MVM_nativecall_make_cppstruct(tc, typeobj, cobj);
+                        }
                         else if(type == MVM_CUNION_ATTR_CUNION) {
                             if (repr_data->attribute_locations[slot] & MVM_CUNION_ATTR_INLINED)
                                 obj = MVM_nativecall_make_cunion(tc, typeobj,
@@ -479,7 +505,7 @@ static void bind_attribute(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
     MVMint64 slot;
 
     if (!repr_data)
-        MVM_exception_throw_adhoc(tc, "P6opaque: must compose before using bind_attribute");
+        MVM_exception_throw_adhoc(tc, "CUnion: must compose before using bind_attribute");
 
     slot = hint >= 0 ? hint : try_get_slot(tc, repr_data, class_handle, name);
     if (slot >= 0) {
@@ -614,6 +640,11 @@ static void gc_mark_repr_data(MVMThreadContext *tc, MVMSTable *st, MVMGCWorklist
     }
 }
 
+/* Free representation data. */
+static void gc_free_repr_data(MVMThreadContext *tc, MVMSTable *st) {
+    MVM_free(st->REPR_data);
+}
+
 /* This is called to do any cleanup of resources when an object gets
  * embedded inside another one. Never called on a top-level object. */
 static void gc_cleanup(MVMThreadContext *tc, MVMSTable *st, void *data) {
@@ -703,7 +734,7 @@ static void deserialize_repr_data(MVMThreadContext *tc, MVMSTable *st, MVMSerial
         repr_data->struct_offsets[i] = MVM_serialization_read_varint(tc, reader);
 
         if(MVM_serialization_read_varint(tc, reader)){
-            repr_data->flattened_stables[i] = MVM_serialization_read_stable_ref(tc, reader);
+            MVM_ASSIGN_REF(tc, &(st->header), repr_data->flattened_stables[i], MVM_serialization_read_stable_ref(tc, reader));
         }
         else {
             repr_data->flattened_stables[i] = NULL;
@@ -766,7 +797,7 @@ static const MVMREPROps this_repr = {
     gc_free,
     gc_cleanup,
     gc_mark_repr_data,
-    NULL, /* gc_free_repr_data */
+    gc_free_repr_data,
     compose,
     NULL, /* spesh */
     "CUnion", /* name */
