@@ -2,9 +2,9 @@
 
 /* Some constants. */
 #define HEADER_SIZE                 92
-#define MIN_BYTECODE_VERSION        2
-#define MAX_BYTECODE_VERSION        4
-#define FRAME_HEADER_SIZE           ((cu->body.bytecode_version >= 4 ? 11 : 9) * 4 + (cu->body.bytecode_version >= 4 ? 3 : 2) * 2)
+#define MIN_BYTECODE_VERSION        5
+#define MAX_BYTECODE_VERSION        5
+#define FRAME_HEADER_SIZE           (11 * 4 + 3 * 2)
 #define FRAME_HANDLER_SIZE          (4 * 4 + 2 * 2)
 #define FRAME_SLV_SIZE              (2 * 2 + 2 * 4)
 #define SCDEP_HEADER_OFFSET         12
@@ -32,16 +32,16 @@ typedef struct {
     MVMuint32  expected_strings;
 
     /* The SC dependencies segment. */
-    MVMuint8  *sc_seg;
     MVMuint32  expected_scs;
+    MVMuint8  *sc_seg;
 
     /* The extension ops segment. */
     MVMuint8 *extop_seg;
     MVMuint32 expected_extops;
 
     /* The frame segment. */
-    MVMuint8  *frame_seg;
     MVMuint32  expected_frames;
+    MVMuint8  *frame_seg;
     MVMuint16 *frame_outer_fixups;
 
     /* The callsites segment. */
@@ -49,8 +49,8 @@ typedef struct {
     MVMuint32  expected_callsites;
 
     /* The bytecode segment. */
-    MVMuint8  *bytecode_seg;
     MVMuint32  bytecode_size;
+    MVMuint8  *bytecode_seg;
 
     /* The annotations segment */
     MVMuint8  *annotation_seg;
@@ -59,6 +59,9 @@ typedef struct {
     /* HLL name string index */
     MVMuint32  hll_str_idx;
 
+    /* The limit we can not read beyond. */
+    MVMuint8 *read_limit;
+
     /* Special frame indexes */
     MVMuint32  main_frame;
     MVMuint32  load_frame;
@@ -108,11 +111,9 @@ static void cleanup_all(MVMThreadContext *tc, ReaderState *rs) {
 
 /* Ensures we can read a certain amount of bytes without overrunning the end
  * of the stream. */
-static void ensure_can_read(MVMThreadContext *tc, MVMCompUnit *cu, ReaderState *rs, MVMuint8 *pos, MVMuint32 size) {
-    MVMCompUnitBody *cu_body = &cu->body;
-    if (pos + size > cu_body->data_start + cu_body->data_size) {
-        if (rs)
-            cleanup_all(tc, rs);
+MVM_STATIC_INLINE void ensure_can_read(MVMThreadContext *tc, MVMCompUnit *cu, ReaderState *rs, MVMuint8 *pos, MVMuint32 size) {
+    if (pos + size > rs->read_limit) {
+        cleanup_all(tc, rs);
         MVM_exception_throw_adhoc(tc, "Read past end of bytecode stream");
     }
 }
@@ -152,6 +153,7 @@ static ReaderState * dissect_bytecode(MVMThreadContext *tc, MVMCompUnit *cu) {
     rs = MVM_malloc(sizeof(ReaderState));
     memset(rs, 0, sizeof(ReaderState));
     rs->version = version;
+    rs->read_limit = cu_body->data_start + cu_body->data_size;
     cu->body.bytecode_version = version;
 
     /* Locate SC dependencies segment. */
@@ -262,15 +264,26 @@ static MVMString ** deserialize_strings(MVMThreadContext *tc, MVMCompUnit *cu, R
     /* Load strings. */
     pos = rs->string_seg;
     for (i = 0; i < rs->expected_strings; i++) {
+        MVMint32 decode_utf8 = 1;
+
         /* Ensure we can read at least a string size here and do so. */
         ensure_can_read(tc, cu, rs, pos, 4);
         ss = read_int32(pos, 0);
         pos += 4;
 
+        /* At high enough bytecode file versions, the LSB on this number
+         * is 1 if we should decode as UTF-8, and 0 if Latin-1 will do. */
+        if (rs->version >= 5) {
+            decode_utf8 = ss & 1;
+            ss = ss >> 1;
+        }
+
         /* Ensure we can read in the string of this size, and decode
          * it if so. */
         ensure_can_read(tc, cu, rs, pos, ss);
-        MVM_ASSIGN_REF(tc, &(cu->common.header), strings[i], MVM_string_utf8_decode(tc, tc->instance->VMString, (char *)pos, ss));
+        MVM_ASSIGN_REF(tc, &(cu->common.header), strings[i], decode_utf8
+            ? MVM_string_utf8_decode(tc, tc->instance->VMString, (char *)pos, ss)
+            : MVM_string_latin1_decode(tc, tc->instance->VMString, (char *)pos, ss));
         pos += ss;
 
         /* Add alignment. */
@@ -560,7 +573,7 @@ static MVMStaticFrame ** deserialize_frames(MVMThreadContext *tc, MVMCompUnit *c
         {
             MVMuint32 skip = 2 * static_frame_body->num_locals +
                              6 * static_frame_body->num_lexicals;
-            MVMuint16 slvs = cu->body.bytecode_version >= 4 ? read_int16(pos, 40) : 0;
+            MVMuint16 slvs = read_int16(pos, 40);
             pos += FRAME_HEADER_SIZE;
             ensure_can_read(tc, cu, rs, pos, skip);
             pos += skip;
@@ -620,7 +633,7 @@ void MVM_bytecode_finish_frame(MVMThreadContext *tc, MVMCompUnit *cu,
     pos = sf->body.frame_data_pos;
 
     /* Get the number of static lex values we'll need to apply. */
-    slvs = cu->body.bytecode_version >= 4 ? read_int16(pos, 40) : 0;
+    slvs = read_int16(pos, 40);
 
     /* Skip past header. */
     pos += FRAME_HEADER_SIZE;
@@ -696,8 +709,10 @@ void MVM_bytecode_finish_frame(MVMThreadContext *tc, MVMCompUnit *cu,
             /* State variable; need to resolve wval immediately. Other kinds
              * can wait. */
             MVMSerializationContext *sc = MVM_sc_get_sc(tc, cu, read_int32(pos, 4));
-            if (sc == NULL)
+            if (sc == NULL) {
+                MVM_reentrantmutex_unlock(tc, (MVMReentrantMutex *)cu->body.update_mutex);
                 MVM_exception_throw_adhoc(tc, "SC not yet resolved; lookup failed");
+            }
             MVM_ASSIGN_REF(tc, &(sf->common.header), sf->body.static_env[lex_idx].o,
                 MVM_sc_get_object(tc, sc, read_int32(pos, 8)));
         }
