@@ -5,36 +5,45 @@
  * bytecode. */
 static void prepare_and_verify_static_frame(MVMThreadContext *tc, MVMStaticFrame *static_frame) {
     MVMStaticFrameBody *static_frame_body = &static_frame->body;
+    MVMCompUnit        *cu                = static_frame_body->cu;
 
     /* Ensure the frame is fully deserialized. */
     if (!static_frame_body->fully_deserialized)
-        MVM_bytecode_finish_frame(tc, static_frame_body->cu, static_frame, 0);
+        MVM_bytecode_finish_frame(tc, cu, static_frame, 0);
 
-    /* Work size is number of locals/registers plus size of the maximum
-     * call site argument list. */
-    static_frame_body->work_size = sizeof(MVMRegister) *
-        (static_frame_body->num_locals + static_frame_body->cu->body.max_callsite_size);
+    /* Take compilation unit lock, to make sure we don't race to do the
+     * frame preparation/verification work. */
+    MVM_reentrantmutex_lock(tc, (MVMReentrantMutex *)cu->body.update_mutex);
+    if (static_frame->body.instrumentation_level == 0) {
+        /* Work size is number of locals/registers plus size of the maximum
+        * call site argument list. */
+        static_frame_body->work_size = sizeof(MVMRegister) *
+            (static_frame_body->num_locals + static_frame_body->cu->body.max_callsite_size);
 
-    /* Validate the bytecode. */
-    MVM_validate_static_frame(tc, static_frame);
+        /* Validate the bytecode. */
+        MVM_validate_static_frame(tc, static_frame);
 
-    /* Obtain an index to each threadcontext's lexotic pool table */
-    static_frame_body->pool_index = MVM_incr(&tc->instance->num_frames_run);
+        /* Obtain an index to each threadcontext's lexotic pool table */
+        static_frame_body->pool_index = MVM_incr(&tc->instance->num_frames_run);
 
-    /* Check if we have any state var lexicals. */
-    if (static_frame_body->static_env_flags) {
-        MVMuint8 *flags  = static_frame_body->static_env_flags;
-        MVMint64  numlex = static_frame_body->num_lexicals;
-        MVMint64  i;
-        for (i = 0; i < numlex; i++)
-            if (flags[i] == 2) {
-                static_frame_body->has_state_vars = 1;
-                break;
-            }
+        /* Check if we have any state var lexicals. */
+        if (static_frame_body->static_env_flags) {
+            MVMuint8 *flags  = static_frame_body->static_env_flags;
+            MVMint64  numlex = static_frame_body->num_lexicals;
+            MVMint64  i;
+            for (i = 0; i < numlex; i++)
+                if (flags[i] == 2) {
+                    static_frame_body->has_state_vars = 1;
+                    break;
+                }
+        }
+
+        /* Set its spesh threshold. */
+        static_frame_body->spesh_threshold = MVM_spesh_threshold(tc, static_frame);
     }
 
-    /* Set its spesh threshold. */
-    static_frame_body->spesh_threshold = MVM_spesh_threshold(tc, static_frame);
+    /* Unlock, now we're finished. */
+    MVM_reentrantmutex_unlock(tc, (MVMReentrantMutex *)cu->body.update_mutex);
 }
 
 /* When we don't match the current instrumentation level, we hit this. It may
@@ -52,40 +61,80 @@ static void instrumentation_level_barrier(MVMThreadContext *tc, MVMStaticFrame *
     /* Add profiling instrumentation if needed. */
     if (tc->instance->profiling)
         MVM_profile_instrument(tc, static_frame);
+    else if (tc->instance->cross_thread_write_logging)
+        MVM_cross_thread_write_instrument(tc, static_frame);
     else
         MVM_profile_ensure_uninstrumented(tc, static_frame);
 }
 
-/* Increases the reference count of a frame. */
+/* Increases the reference count of a frame, and marks it as having been
+ * referenced by a garbage collectable object. Can only safely be used
+ * if you know the frame is referenced by at least one other frame or by
+ * being on the stack top. If there's a chance it may be referenced only
+ * by an object, use MVM_frame_acquire_ref to avoid races. */
 MVMFrame * MVM_frame_inc_ref(MVMThreadContext *tc, MVMFrame *frame) {
+    frame->refd_by_object = 1;
     MVM_incr(&frame->ref_count);
     return frame;
 }
 
+/* Increases the reference count of a frame. Can only safely be used when
+ * the thing that will refer to the frame is another frame, and when you
+ * know the frame is referenced by at least one other frame or by being on
+ * the stack top. If there's a chance it may be referenced only by an object,
+ * use MVM_frame_acquire_ref to avoid races. */
+MVMFrame * MVM_frame_inc_ref_by_frame(MVMThreadContext *tc, MVMFrame *frame) {
+    MVM_incr(&frame->ref_count);
+    return frame;
+}
+
+/* Takes a pointer to a location that points to an MVMFrame, and safely
+ * obtains a reference to the frame. If the frame should die while we
+ * try to do so, then it will dereference the frame pointer afresh and
+ * try again. */
+MVMFrame * MVM_frame_acquire_ref(MVMThreadContext *tc, MVMFrame **frame) {
+    while (1) {
+        MVMFrame *try = (MVMFrame *)MVM_load(frame);
+        if (try) {
+            AO_t ref_count = MVM_load(&(try->ref_count));
+            if (ref_count > 0)
+                if (MVM_trycas(&(try->ref_count), ref_count, ref_count + 1))
+                    return try;
+        }
+        else {
+            return NULL;
+        }
+    }
+}
+
 /* Decreases the reference count of a frame. If it hits zero, then we can
  * free it. Returns null for convenience. */
 MVMFrame * MVM_frame_dec_ref(MVMThreadContext *tc, MVMFrame *frame) {
     /* MVM_dec returns what the count was before it decremented it
      * to zero, so we look for 1 here. */
     while (MVM_decr(&frame->ref_count) == 1) {
-        MVMuint32 pool_index = frame->static_info->body.pool_index;
         MVMFrame *outer_to_decr = frame->outer;
 
         /* If there's a caller pointer, decrement that. */
         if (frame->caller)
             frame->caller = MVM_frame_dec_ref(tc, frame->caller);
 
-        /* Destroy the frame. */
-        if (frame->env) {
-            MVM_fixed_size_free(tc, tc->instance->fsa, frame->allocd_env,
-                frame->env);
-        }
+        /* Destroy the frame. If it was referenced by an object, we need to
+         * free it at the next safe point; the 0'd ref count must remain
+         * readable until then since we use it as a sentinel value to avoid
+         * anything acquiring a reference to a dead frame. */
         if (frame->work) {
             MVM_args_proc_cleanup(tc, &frame->params);
             MVM_fixed_size_free(tc, tc->instance->fsa, frame->allocd_work,
                 frame->work);
         }
-        MVM_fixed_size_free(tc, tc->instance->fsa, sizeof(MVMFrame), frame);
+        if (frame->env)
+            MVM_fixed_size_free_at_safepoint(tc, tc->instance->fsa,
+                frame->allocd_env, frame->env);
+        if (frame->refd_by_object)
+            MVM_fixed_size_free_at_safepoint(tc, tc->instance->fsa, sizeof(MVMFrame), frame);
+        else
+            MVM_fixed_size_free(tc, tc->instance->fsa, sizeof(MVMFrame), frame);
 
         /* Decrement any outer. */
         if (outer_to_decr)
@@ -155,8 +204,9 @@ static MVMFrame * create_context_only(MVMThreadContext *tc, MVMStaticFrame *stat
  * with a ref count of 1 due to being held by an SC. */
 MVMFrame * MVM_frame_create_context_only(MVMThreadContext *tc, MVMStaticFrame *static_frame,
         MVMObject *code_ref) {
-    return MVM_frame_inc_ref(tc,
-        create_context_only(tc, static_frame, code_ref, 0));
+    MVMFrame *f = create_context_only(tc, static_frame, code_ref, 0);
+    f->ref_count = 1;
+    return f;
 }
 
 /* Provides auto-close functionality, for the handful of cases where we have
@@ -181,11 +231,11 @@ static MVMFrame * autoclose(MVMThreadContext *tc, MVMStaticFrame *needed) {
         if (outer_code->body.outer &&
                 outer_code->body.outer->static_info->body.bytecode == needed->body.bytecode) {
             /* Yes, just take it. */
-            result->outer = MVM_frame_inc_ref(tc, outer_code->body.outer);
+            result->outer = MVM_frame_acquire_ref(tc, &(outer_code->body.outer));
         }
         else {
             /* Otherwise, recursively auto-close. */
-            result->outer = MVM_frame_inc_ref(tc, autoclose(tc, needed->body.outer));
+            result->outer = MVM_frame_inc_ref_by_frame(tc, autoclose(tc, needed->body.outer));
         }
     }
     return result;
@@ -195,12 +245,12 @@ static MVMFrame * autoclose(MVMThreadContext *tc, MVMStaticFrame *needed) {
 static MVMFrame * allocate_frame(MVMThreadContext *tc, MVMStaticFrameBody *static_frame_body,
                                  MVMSpeshCandidate *spesh_cand) {
     MVMFrame *frame = NULL;
-    MVMFrame *node;
     MVMint32  env_size, work_size;
 
     /* Allocate the frame. */
     frame = MVM_fixed_size_alloc(tc, tc->instance->fsa, sizeof(MVMFrame));
     frame->params.named_used = NULL;
+    frame->refd_by_object = 0;
 
     /* Ensure special return pointers, continuation tags, dynlex cache,
      * and return address are null. */
@@ -247,7 +297,8 @@ static MVMFrame * allocate_frame(MVMThreadContext *tc, MVMStaticFrameBody *stati
 void MVM_frame_invoke_code(MVMThreadContext *tc, MVMCode *code,
                            MVMCallsite *callsite, MVMint32 spesh_cand) {
     MVM_frame_invoke(tc, code->body.sf, callsite,  tc->cur_frame->args,
-                     code->body.outer, (MVMObject*)code, spesh_cand);
+        MVM_frame_acquire_ref(tc, &(code->body.outer)),
+        (MVMObject*)code, spesh_cand);
 }
 
 /* Takes a static frame and a thread context. Invokes the static frame. */
@@ -392,42 +443,83 @@ void MVM_frame_invoke(MVMThreadContext *tc, MVMStaticFrame *static_frame,
 
     /* Outer. */
     if (outer) {
-        /* We were provided with an outer frame; just ensure that it is
-         * based on the correct static frame (compare on bytecode address
-         * to cope with nqp::freshcoderef). */
+        /* We were provided with an outer frame and it will already have had
+         * its reference count incremented; just ensure that it is based on the
+         * correct static frame (compare on bytecode address to cope with
+         * nqp::freshcoderef). */
         if (outer->static_info->body.orig_bytecode == static_frame_body->outer->body.orig_bytecode)
             frame->outer = outer;
-        else
-            MVM_exception_throw_adhoc(tc,
+        else {
+            char *frame_cuuid = MVM_string_utf8_encode_C_string(tc, static_frame_body->cuuid);
+            char *frame_name;
+            char *outer_cuuid = MVM_string_utf8_encode_C_string(tc, outer->static_info->body.cuuid);
+            char *outer_name;
+            char *frame_outer_cuuid = MVM_string_utf8_encode_C_string(tc, static_frame_body->outer->body.cuuid);
+            char *frame_outer_name;
+
+            char *waste[7] = { frame_cuuid, outer_cuuid, frame_outer_cuuid, NULL, NULL, NULL, NULL };
+            int waste_counter = 3;
+
+            if (static_frame_body->name) {
+                frame_name = MVM_string_utf8_encode_C_string(tc, static_frame_body->name);
+                waste[waste_counter++] = frame_name;
+            }
+            else {
+                frame_name = "<anonymous static frame>";
+            }
+
+            if (outer->static_info->body.name) {
+                outer_name = MVM_string_utf8_encode_C_string(tc, outer->static_info->body.name);
+                waste[waste_counter++] = outer_name;
+            }
+            else {
+                outer_name = "<anonymous static frame>";
+            }
+
+            if (static_frame_body->outer->body.name) {
+                frame_outer_name = MVM_string_utf8_encode_C_string(tc, static_frame_body->outer->body.name);
+                waste[waste_counter++] = frame_outer_name;
+            }
+            else {
+                frame_outer_name = "<anonymous static frame>";
+            }
+
+            MVM_exception_throw_adhoc_free(tc, waste,
                 "When invoking %s '%s', provided outer frame %p (%s '%s') does not match expected static frame %p (%s '%s')",
-                MVM_string_utf8_encode_C_string(tc, static_frame_body->cuuid),
-                static_frame_body->name ? MVM_string_utf8_encode_C_string(tc, static_frame_body->name) : "<anonymous static frame>",
+                frame_cuuid,
+                frame_name,
                 outer->static_info,
-                MVM_string_utf8_encode_C_string(tc, outer->static_info->body.cuuid),
-                outer->static_info->body.name ? MVM_string_utf8_encode_C_string(tc, outer->static_info->body.name) : "<anonymous static frame>",
+                outer_cuuid,
+                outer_name,
                 static_frame_body->outer,
-                MVM_string_utf8_encode_C_string(tc, static_frame_body->outer->body.cuuid),
-                static_frame_body->outer->body.name ? MVM_string_utf8_encode_C_string(tc, static_frame_body->outer->body.name) : "<anonymous static frame>");
+                frame_outer_cuuid,
+                frame_outer_name);
+        }
     }
-    else if (static_frame_body->static_code && static_frame_body->static_code->body.outer) {
-        /* We're lacking an outer, but our static code object may have one.
-         * This comes up in the case of cloned protoregexes, for example. */
-        frame->outer = static_frame_body->static_code->body.outer;
-    }
-    else if (static_frame_body->outer) {
-        /* Auto-close, and cache it in the static frame. */
-        frame->outer = autoclose(tc, static_frame_body->outer);
-        static_frame_body->static_code->body.outer = MVM_frame_inc_ref(tc, frame->outer);
+    else if (static_frame_body->static_code) {
+        MVMCode *static_code = static_frame_body->static_code;
+        if (static_code->body.outer) {
+            /* We're lacking an outer, but our static code object may have one.
+            * This comes up in the case of cloned protoregexes, for example. */
+            frame->outer = MVM_frame_inc_ref_by_frame(tc, static_code->body.outer);
+        }
+        else if (static_frame_body->outer) {
+            /* Auto-close, and cache it in the static frame. */
+            frame->outer = MVM_frame_inc_ref_by_frame(tc,
+                autoclose(tc, static_frame_body->outer));
+            static_code->body.outer = MVM_frame_inc_ref(tc, frame->outer);
+        }
+        else {
+            frame->outer = NULL;
+        }
     }
     else {
         frame->outer = NULL;
     }
-    if (frame->outer)
-        MVM_frame_inc_ref(tc, frame->outer);
 
     /* Caller is current frame in the thread context. */
     if (tc->cur_frame)
-        frame->caller = MVM_frame_inc_ref(tc, tc->cur_frame);
+        frame->caller = MVM_frame_inc_ref_by_frame(tc, tc->cur_frame);
     else
         frame->caller = NULL;
     frame->keep_caller = 0;
@@ -533,7 +625,7 @@ MVMFrame * MVM_frame_create_for_deopt(MVMThreadContext *tc, MVMStaticFrame *stat
     frame->params.arg_flags         = NULL;
     frame->params.named_used        = NULL;
     if (code_ref->body.outer)
-        frame->outer = MVM_frame_inc_ref(tc, code_ref->body.outer);
+        frame->outer = MVM_frame_acquire_ref(tc, &(code_ref->body.outer));
     else
         frame->outer = NULL;
     return frame;
@@ -918,18 +1010,23 @@ MVMRegister * MVM_frame_find_lexical_by_name(MVMThreadContext *tc, MVMString *na
                     return result;
                 }
                 else {
-                   MVM_exception_throw_adhoc(tc,
+                    char *c_name = MVM_string_utf8_encode_C_string(tc, name);
+                    char *waste[] = { c_name, NULL };
+                    MVM_exception_throw_adhoc_free(tc, waste,
                         "Lexical with name '%s' has wrong type",
-                            MVM_string_utf8_encode_C_string(tc, name));
+                            c_name);
                 }
             }
         }
         cur_frame = cur_frame->outer;
     }
-    if (type == MVM_reg_obj)
-        return NULL;
-    MVM_exception_throw_adhoc(tc, "No lexical found with name '%s'",
-        MVM_string_utf8_encode_C_string(tc, name));
+    if (type != MVM_reg_obj) {
+        char *c_name = MVM_string_utf8_encode_C_string(tc, name);
+        char *waste[] = { c_name, NULL };
+        MVM_exception_throw_adhoc_free(tc, waste, "No lexical found with name '%s'",
+            c_name);
+    }
+    return NULL;
 }
 
 /* Finds a lexical in the outer frame, throwing if it's not there. */
@@ -937,9 +1034,12 @@ MVMObject * MVM_frame_find_lexical_by_name_outer(MVMThreadContext *tc, MVMString
     MVMRegister *r = MVM_frame_find_lexical_by_name_rel(tc, name, tc->cur_frame->outer);
     if (r)
         return r->o;
-    else
-        MVM_exception_throw_adhoc(tc, "No lexical found with name '%s'",
-            MVM_string_utf8_encode_C_string(tc, name));
+    else {
+        char *c_name = MVM_string_utf8_encode_C_string(tc, name);
+        char *waste[] = { c_name, NULL };
+        MVM_exception_throw_adhoc_free(tc, waste, "No lexical found with name '%s'",
+            c_name);
+    }
 }
 
 /* Looks up the address of the lexical with the specified name, starting with
@@ -962,9 +1062,11 @@ MVMRegister * MVM_frame_find_lexical_by_name_rel(MVMThreadContext *tc, MVMString
                     return result;
                 }
                 else {
-                   MVM_exception_throw_adhoc(tc,
+                    char *c_name = MVM_string_utf8_encode_C_string(tc, name);
+                    char *waste[] = { c_name, NULL };
+                    MVM_exception_throw_adhoc_free(tc, waste,
                         "Lexical with name '%s' has wrong type",
-                            MVM_string_utf8_encode_C_string(tc, name));
+                            c_name);
                 }
             }
         }
@@ -995,9 +1097,11 @@ MVMRegister * MVM_frame_find_lexical_by_name_rel_caller(MVMThreadContext *tc, MV
                         return result;
                     }
                     else {
-                       MVM_exception_throw_adhoc(tc,
+                        char *c_name = MVM_string_utf8_encode_C_string(tc, name);
+                        char *waste[] = { c_name, NULL };
+                        MVM_exception_throw_adhoc_free(tc, waste,
                             "Lexical with name '%s' has wrong type",
-                                MVM_string_utf8_encode_C_string(tc, name));
+                                c_name);
                     }
                 }
             }
@@ -1229,8 +1333,10 @@ void MVM_frame_binddynlex(MVMThreadContext *tc, MVMString *name, MVMObject *valu
     MVMuint16 type;
     MVMRegister *lex_reg = MVM_frame_find_contextual_by_name(tc, name, &type, cur_frame, 0);
     if (!lex_reg) {
-        MVM_exception_throw_adhoc(tc, "No contextual found with name '%s'",
-            MVM_string_utf8_encode_C_string(tc, name));
+        char *c_name = MVM_string_utf8_encode_C_string(tc, name);
+        char *waste[] = { c_name, NULL };
+        MVM_exception_throw_adhoc_free(tc, waste, "No contextual found with name '%s'",
+            c_name);
     }
     switch (type) {
         case MVM_reg_int64:
@@ -1264,8 +1370,13 @@ MVMRegister * MVM_frame_lexical(MVMThreadContext *tc, MVMFrame *f, MVMString *na
         if (entry)
             return &f->env[entry->value];
     }
-    MVM_exception_throw_adhoc(tc, "Frame has no lexical with name '%s'",
-        MVM_string_utf8_encode_C_string(tc, name));
+
+    {
+        char *c_name = MVM_string_utf8_encode_C_string(tc, name);
+        char *waste[] = { c_name, NULL };
+        MVM_exception_throw_adhoc_free(tc, waste, "Frame has no lexical with name '%s'",
+            c_name);
+    }
 }
 
 /* Returns the storage unit for the lexical in the specified frame. */
@@ -1303,19 +1414,29 @@ MVMuint16 MVM_frame_lexical_primspec(MVMThreadContext *tc, MVMFrame *f, MVMStrin
                 case MVM_reg_obj:
                     return MVM_STORAGE_SPEC_BP_NONE;
                 default:
-                    MVM_exception_throw_adhoc(tc,
+                {
+                    char *c_name = MVM_string_utf8_encode_C_string(tc, name);
+                    char *waste[] = { c_name, NULL };
+                    MVM_exception_throw_adhoc_free(tc, waste,
                         "Unhandled lexical type in lexprimspec for '%s'",
-                        MVM_string_utf8_encode_C_string(tc, name));
+                        c_name);
+                }
             }
         }
     }
-    MVM_exception_throw_adhoc(tc, "Frame has no lexical with name '%s'",
-        MVM_string_utf8_encode_C_string(tc, name));
+    {
+        char *c_name = MVM_string_utf8_encode_C_string(tc, name);
+        char *waste[] = { c_name, NULL };
+        MVM_exception_throw_adhoc_free(tc, waste, "Frame has no lexical with name '%s'",
+            c_name);
+    }
 }
 
 static MVMObject * find_invokee_internal(MVMThreadContext *tc, MVMObject *code, MVMCallsite **tweak_cs, MVMInvocationSpec *is) {
     if (!MVM_is_null(tc, is->class_handle)) {
         MVMRegister dest;
+        if (!IS_CONCRETE(code))
+            MVM_exception_throw_adhoc(tc, "Can not invoke a code type object");
         REPR(code)->attr_funcs.get_attribute(tc,
             STABLE(code), code, OBJECT_BODY(code),
             is->class_handle, is->attr_name,
@@ -1378,13 +1499,14 @@ MVMObject * MVM_frame_find_invokee_multi_ok(MVMThreadContext *tc, MVMObject *cod
     if (STABLE(code)->invoke == MVM_6model_invoke_default) {
         MVMInvocationSpec *is = STABLE(code)->invocation_spec;
         if (!is) {
-            MVM_exception_throw_adhoc(tc, "Cannot invoke this object (REPR: %s, cs = %d)",
-                REPR(code)->name, STABLE(code)->container_spec ? 1 : 0);
+            MVM_exception_throw_adhoc(tc, "Cannot invoke this object");
         }
         if (!MVM_is_null(tc, is->md_class_handle)) {
             /* We might be able to dig straight into the multi cache and not
              * have to invoke the proto. */
             MVMRegister dest;
+            if (!IS_CONCRETE(code))
+                MVM_exception_throw_adhoc(tc, "Can not invoke a code type object");
             REPR(code)->attr_funcs.get_attribute(tc,
                 STABLE(code), code, OBJECT_BODY(code),
                 is->md_class_handle, is->md_valid_attr_name,
