@@ -317,6 +317,10 @@ static void write_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
     wi->buf           = uv_buf_init(output, output_size);
     wi->req->data     = data;
     handle_data       = (MVMIOAsyncSocketData *)wi->handle->body.data;
+
+    if (uv_is_closing((uv_handle_t *)handle_data->handle))
+        MVM_exception_throw_adhoc(tc, "cannot write to a closed socket");
+
     if ((r = uv_write(wi->req, handle_data->handle, &(wi->buf), 1, on_write)) < 0) {
         /* Error; need to notify. */
         MVMROOT(tc, async_task, {
@@ -446,7 +450,12 @@ static void close_cb(uv_handle_t *handle) {
     MVM_free(handle);
 }
 static void close_perform(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_task, void *data) {
-    uv_close((uv_handle_t *)data, close_cb);
+    uv_handle_t *handle = (uv_handle_t *)data;
+
+    if (uv_is_closing(handle))
+        MVM_exception_throw_adhoc(tc, "cannot close a closed socket");
+
+    uv_close(handle, close_cb);
 }
 
 /* Operations table for async close task. */
@@ -459,17 +468,16 @@ static const MVMAsyncTaskOps close_op_table = {
 
 static MVMint64 close_socket(MVMThreadContext *tc, MVMOSHandle *h) {
     MVMIOAsyncSocketData *data = (MVMIOAsyncSocketData *)h->body.data;
-    if (data->handle) {
-        MVMAsyncTask *task;
-        MVMROOT(tc, h, {
-            task = (MVMAsyncTask *)MVM_repr_alloc_init(tc,
-                tc->instance->boot_types.BOOTAsync);
-        });
-        task->body.ops  = &close_op_table;
-        task->body.data = data->handle;
-        MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
-        data->handle = NULL;
-    }
+    MVMAsyncTask *task;
+
+    MVMROOT(tc, h, {
+        task = (MVMAsyncTask *)MVM_repr_alloc_init(tc,
+            tc->instance->boot_types.BOOTAsync);
+    });
+    task->body.ops  = &close_op_table;
+    task->body.data = data->handle;
+    MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+
     return 0;
 }
 
