@@ -2,8 +2,6 @@
 #include "platform/time.h"
 #include "tinymt64.h"
 
-#include <math.h>
-
 /* concatenating with "" ensures that only literal strings are accepted as argument. */
 #define STR_WITH_LEN(str)  ("" str ""), (sizeof(str) - 1)
 
@@ -576,6 +574,7 @@ static MVMint64 close_stdin(MVMThreadContext *tc, MVMOSHandle *h) {
         task->body.ops  = &close_op_table;
         task->body.data = si->stdin_handle;
         MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
+        si->stdin_handle = NULL;
     }
     return 0;
 }
@@ -594,7 +593,7 @@ static const MVMIOOps proc_op_table = {
     NULL,
     NULL,
     NULL,
-    &proc_async_gc_mark,
+    proc_async_gc_mark,
     NULL
 };
 
@@ -608,6 +607,7 @@ static void async_spawn_on_exit(uv_process_t *req, MVMint64 exit_status, int ter
     MVMThreadContext *tc  = si->tc;
     MVMObject *done_cb = MVM_repr_at_key_o(tc, si->callbacks,
         tc->instance->str_consts.done);
+    MVMOSHandle *os_handle;
     if (!MVM_is_null(tc, done_cb)) {
         MVMROOT(tc, done_cb, {
             /* Get status. */
@@ -631,6 +631,12 @@ static void async_spawn_on_exit(uv_process_t *req, MVMint64 exit_status, int ter
         });
     }
 
+    /* when invoked via MVMIOOps, close_stdin is already wrapped in a mutex */
+    os_handle = (MVMOSHandle *) si->handle;
+    uv_mutex_lock(os_handle->body.mutex);
+    close_stdin(tc, os_handle);
+    uv_mutex_unlock(os_handle->body.mutex);
+
     /* Close handle. */
     uv_close((uv_handle_t *)req, spawn_async_close);
     ((MVMIOAsyncProcessData *)((MVMOSHandle *)si->handle)->body.data)->handle = NULL;
@@ -780,7 +786,7 @@ static void spawn_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
         process_stdio[1].data.stream = (uv_stream_t *)pipe;
         si->ds_stdout                = MVM_string_decodestream_create(tc, MVM_encoding_type_utf8, 0);
         stdout_pipe                  = pipe;
-        stdout_cb                    = &async_spawn_stdout_chars_read;
+        stdout_cb                    = async_spawn_stdout_chars_read;
     }
     else if (MVM_repr_exists_key(tc, si->callbacks, tc->instance->str_consts.stdout_bytes)) {
         uv_pipe_t *pipe = MVM_malloc(sizeof(uv_pipe_t));
@@ -789,7 +795,7 @@ static void spawn_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
         process_stdio[1].flags       = UV_CREATE_PIPE | UV_WRITABLE_PIPE;
         process_stdio[1].data.stream = (uv_stream_t *)pipe;
         stdout_pipe                  = pipe;
-        stdout_cb                    = &async_spawn_stdout_bytes_read;
+        stdout_cb                    = async_spawn_stdout_bytes_read;
     }
     else {
         process_stdio[1].flags   = UV_INHERIT_FD;
@@ -803,7 +809,7 @@ static void spawn_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
         process_stdio[2].data.stream = (uv_stream_t *)pipe;
         si->ds_stderr                = MVM_string_decodestream_create(tc, MVM_encoding_type_utf8, 0);
         stderr_pipe                  = pipe;
-        stderr_cb                    = &async_spawn_stderr_chars_read;
+        stderr_cb                    = async_spawn_stderr_chars_read;
     }
     else if (MVM_repr_exists_key(tc, si->callbacks, tc->instance->str_consts.stderr_bytes)) {
         uv_pipe_t *pipe = MVM_malloc(sizeof(uv_pipe_t));
@@ -812,7 +818,7 @@ static void spawn_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
         process_stdio[2].flags       = UV_CREATE_PIPE | UV_WRITABLE_PIPE;
         process_stdio[2].data.stream = (uv_stream_t *)pipe;
         stderr_pipe                  = pipe;
-        stderr_cb                    = &async_spawn_stderr_bytes_read;
+        stderr_cb                    = async_spawn_stderr_bytes_read;
     }
     else {
         process_stdio[2].flags   = UV_INHERIT_FD;
@@ -855,7 +861,21 @@ static void spawn_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
     else {
         MVMOSHandle           *handle  = (MVMOSHandle *)si->handle;
         MVMIOAsyncProcessData *apd     = (MVMIOAsyncProcessData *)handle->body.data;
+        MVMObject *ready_cb;
         apd->handle                    = process;
+
+        ready_cb = MVM_repr_at_key_o(tc, si->callbacks,
+            tc->instance->str_consts.ready);
+
+        if (!MVM_is_null(tc, ready_cb)) {
+            MVMROOT(tc, ready_cb, {
+            MVMROOT(tc, async_task, {
+                MVMObject *arr = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTArray);
+                MVM_repr_push_o(tc, arr, ready_cb);
+                MVM_repr_push_o(tc, ((MVMAsyncTask *)async_task)->body.queue, arr);
+            });
+            });
+        }
     }
 
     /* Start any output readers. */
@@ -919,7 +939,7 @@ static void spawn_gc_free(MVMThreadContext *tc, MVMObject *t, void *data) {
             si->ds_stdout = NULL;
         }
         if (si->ds_stderr) {
-            MVM_string_decodestream_destory(tc, si->ds_stdout);
+            MVM_string_decodestream_destory(tc, si->ds_stderr);
             si->ds_stderr = NULL;
         }
         MVM_free(si);
