@@ -74,6 +74,9 @@ static MVMint32 read_to_buffer(MVMThreadContext *tc, MVMIOSyncStreamData *data,
             MVM_exception_throw_adhoc(tc, "Reading from stream failed: %s",
                 uv_strerror(r));
         uv_ref((uv_handle_t *)data->handle);
+        if (tc->loop != data->handle->loop) {
+            MVM_exception_throw_adhoc(tc, "Tried to read() on a socket from outside its originating thread");
+        }
         uv_run(tc->loop, UV_RUN_DEFAULT);
         return 1;
     }
@@ -242,9 +245,9 @@ static MVMint64 not_std_handle(MVMThreadContext *tc, MVMObject *h) {
 static MVMint64 closefh(MVMThreadContext *tc, MVMOSHandle *h) {
     MVMIOSyncStreamData *data = (MVMIOSyncStreamData *)h->body.data;
     if (data->handle && not_std_handle(tc, (MVMObject *)h)) {
-         uv_close((uv_handle_t *)data->handle, NULL);
-         data->handle = NULL;
-         if (data->ds) {
+        uv_close((uv_handle_t *)data->handle, NULL);
+        data->handle = NULL;
+        if (data->ds) {
             MVM_string_decodestream_destory(tc, data->ds);
             data->ds = NULL;
         }
@@ -252,6 +255,14 @@ static MVMint64 closefh(MVMThreadContext *tc, MVMOSHandle *h) {
     return 0;
 }
 
+/* Operations aiding process spawning and I/O handling. */
+static void bind_stdio_handle(MVMThreadContext *tc, MVMOSHandle *h, uv_stdio_container_t *stdio,
+        uv_process_t *process) {
+    MVMIOSyncStreamData *data = (MVMIOSyncStreamData *)h->body.data;
+    stdio->flags              = UV_INHERIT_STREAM;
+    stdio->data.stream        = data->handle;
+}
+
 /* Frees data associated with the handle, closing it if needed. */
 static void gc_free(MVMThreadContext *tc, MVMObject *h, void *d) {
     MVMIOSyncStreamData *data = (MVMIOSyncStreamData *)d;
@@ -283,6 +294,7 @@ static const MVMIOSyncWritable sync_writable = { MVM_io_syncstream_write_str,
                                                  MVM_io_syncstream_truncate };
 static const MVMIOSeekable          seekable = { MVM_io_syncstream_seek,
                                                  MVM_io_syncstream_tell };
+static const MVMIOPipeable     pipeable      = { bind_stdio_handle };
 static const MVMIOOps op_table = {
     &closable,
     &encodable,
@@ -292,7 +304,7 @@ static const MVMIOOps op_table = {
     NULL,
     &seekable,
     NULL,
-    NULL,
+    &pipeable,
     NULL,
     NULL,
     gc_free
