@@ -808,10 +808,12 @@ static void compile_instruction(VM, WriterState *ws, MASTNode *node) {
         if (ELEMS(vm, o->operands) != num_operands) {
             unsigned int  current_frame_idx = ws->current_frame_idx;
             unsigned int  current_ins_idx = ws->current_ins_idx;
+            char *c_name = VM_STRING_TO_C_STRING(vm, o->name);
+            char *waste[] = { c_name, NULL };
             cleanup_all(vm, ws);
-            DIE(vm, "At Frame %u, Instruction %u, op '%s' has invalid number (%u) of operands; needs %u.",
+            DIE_FREE(vm, waste, "At Frame %u, Instruction %u, op '%s' has invalid number (%u) of operands; needs %u.",
                 current_frame_idx, current_ins_idx,
-                VM_STRING_TO_C_STRING(vm, o->name),
+                c_name,
                 ELEMS(vm, o->operands), num_operands);
         }
 
@@ -1233,9 +1235,9 @@ static void compile_frame(VM, WriterState *ws, MASTNode *node, unsigned short id
     write_int32(ws->frame_seg, fs->frame_start + 34, fs->num_handlers);
 
     /* Write handlers. */
-    ensure_space(vm, &ws->frame_seg, &ws->frame_alloc, ws->frame_pos,
-        FRAME_HANDLER_SIZE * fs->num_handlers);
     for (i = 0; i < fs->num_handlers; i++) {
+        ensure_space(vm, &ws->frame_seg, &ws->frame_alloc, ws->frame_pos,
+            FRAME_HANDLER_SIZE);
         write_int32(ws->frame_seg, ws->frame_pos, fs->handlers[i].start_offset);
         ws->frame_pos += 4;
         write_int32(ws->frame_seg, ws->frame_pos, fs->handlers[i].end_offset);
@@ -1254,6 +1256,7 @@ static void compile_frame(VM, WriterState *ws, MASTNode *node, unsigned short id
             write_int32(ws->frame_seg, ws->frame_pos, 0);
         ws->frame_pos += 4;
         if (fs->handlers[i].category_mask & MVM_EX_CAT_LABELED) {
+            ensure_space(vm, &ws->frame_seg, &ws->frame_alloc, ws->frame_pos, 2);
             write_int16(ws->frame_seg, ws->frame_pos, fs->handlers[i].label_reg);
             ws->frame_pos += 2;
         }
