@@ -496,8 +496,9 @@ MVMObject * MVM_bigint_mod(MVMThreadContext *tc, MVMObject *result_type, MVMObje
         mp_init(ic);
 
         mp_result = mp_mod(ia, ib, ic);
+        clear_temp_bigints(tmp, 2);
+
         if (mp_result == MP_VAL) {
-            clear_temp_bigints(tmp, 2);
             MVM_exception_throw_adhoc(tc, "Division by zero");
         }
         store_bigint_result(bc, ic);
@@ -881,6 +882,9 @@ MVMint64 MVM_bigint_is_prime(MVMThreadContext *tc, MVMObject *a, MVMint64 b) {
     }
 }
 
+/* concatenating with "" ensures that only literal strings are accepted as argument. */
+#define STR_WITH_LEN(str)  ("" str ""), (sizeof(str) - 1)
+
 MVMObject * MVM_bigint_radix(MVMThreadContext *tc, MVMint64 radix, MVMString *str, MVMint64 offset, MVMint64 flag, MVMObject *type) {
     MVMObject *result;
     MVMint64 chars  = MVM_string_graphs(tc, str);
@@ -941,10 +945,25 @@ MVMObject * MVM_bigint_radix(MVMThreadContext *tc, MVMint64 radix, MVMString *st
         ch = (offset < chars) ? MVM_string_get_grapheme_at_nocheck(tc, str, offset) : 0;
     }
 
-   while (offset < chars) {
-        if (ch >= '0' && ch <= '9') ch = ch - '0';
+    while (offset < chars) {
+        if (ch >= '0' && ch <= '9') ch = ch - '0'; /* fast-path for ASCII 0..9 */
         else if (ch >= 'a' && ch <= 'z') ch = ch - 'a' + 10;
         else if (ch >= 'A' && ch <= 'Z') ch = ch - 'A' + 10;
+        else if (ch >= 0xFF21 && ch <= 0xFF3A) ch = ch - 0xFF21 + 10; /* uppercase fullwidth */
+        else if (ch >= 0xFF41 && ch <= 0xFF5A) ch = ch - 0xFF41 + 10; /* lowercase fullwidth */
+        else if (MVM_unicode_codepoint_has_property_value(tc, ch, MVM_UNICODE_PROPERTY_GENERAL_CATEGORY,
+                MVM_unicode_cname_to_property_value_code(tc, MVM_UNICODE_PROPERTY_GENERAL_CATEGORY, STR_WITH_LEN("Nd")))) {
+            /* as of Unicode 6.0.0, characters with the 'de' Numeric Type (and are
+             * thus also of General Category Nd, since 4.0.0) are contiguous
+             * sequences of 10 chars whose Numeric Values ascend from 0 through 9.
+             */
+
+            /* the string returned for NUMERIC_VALUE contains a floating point
+             * value, so atoi will stop on the . in the string. This is fine
+             * though, since we'd have to truncate the float regardless.
+             */
+            ch = atoi(MVM_unicode_codepoint_get_property_cstr(tc, ch, MVM_UNICODE_PROPERTY_NUMERIC_VALUE));
+        }
         else break;
         if (ch >= radix) break;
         mp_mul_d(&zvalue, radix, &zvalue);
