@@ -92,6 +92,7 @@ MVMInstance * MVM_vm_create_instance(void) {
     instance->threads->body.stage = MVM_thread_stage_started;
     instance->threads->body.tc = instance->main_thread;
     instance->threads->body.native_thread_id = MVM_platform_thread_id();
+    instance->threads->body.thread_id = instance->main_thread->thread_id;
 
     /* Create compiler registry */
     instance->compiler_registry = MVM_repr_alloc_init(instance->main_thread, instance->boot_types.BOOTHash);
@@ -228,14 +229,14 @@ void MVM_vm_run_file(MVMInstance *instance, const char *filename) {
 
         /* Run deserialization frame, if there is one. */
         if (cu->body.deserialize_frame) {
-            MVM_interp_run(tc, &toplevel_initial_invoke, cu->body.deserialize_frame);
+            MVM_interp_run(tc, toplevel_initial_invoke, cu->body.deserialize_frame);
         }
     });
 
     /* Run the frame marked main, or if there is none then fall back to the
      * first frame. */
     start_frame = cu->body.main_frame ? cu->body.main_frame : cu->body.frames[0];
-    MVM_interp_run(tc, &toplevel_initial_invoke, start_frame);
+    MVM_interp_run(tc, toplevel_initial_invoke, start_frame);
 }
 
 /* Loads bytecode from the specified file name and dumps it. */
