@@ -331,6 +331,12 @@ static void dump_facts(MVMThreadContext *tc, DumpStr *ds, MVMSpeshGraph *g) {
             if (flags & 2048) {
                 append(ds, " KBxSr");
             }
+            if (flags & 4096) {
+                append(ds, " MgWLG");
+            }
+            if (g->facts[i][j].writer && g->facts[i][j].writer->info->opcode == MVM_SSA_PHI) {
+                appendf(ds, " (merged from %d regs)", g->facts[i][j].writer->info->num_operands - 1);
+            }
             append(ds, "\n");
         }
     }
@@ -371,7 +377,7 @@ static void dump_log_values(MVMThreadContext *tc, DumpStr *ds, MVMSpeshGraph *g)
                 appendf(ds, "% 4d  ", log_obj_idx + 1);
                 interesting = 1;
             } else {
-                appendf(ds, "% 4s  ", "_");
+                appendf(ds, "%4s  ", "_");
             }
 
         }
@@ -384,7 +390,7 @@ static void dump_log_values(MVMThreadContext *tc, DumpStr *ds, MVMSpeshGraph *g)
         appendf(ds, "    %d: %p", log_index + 1, seen_table[log_index]);
         if (STABLE(seen_table[log_index])->REPR->ID == MVM_REPR_ID_P6int) {
             if (IS_CONCRETE(seen_table[log_index]))
-                appendf(ds, " P6int(%d)", MVM_repr_get_int(tc, (MVMObject*)seen_table[log_index]));
+                appendf(ds, " P6int(%"PRId64")", MVM_repr_get_int(tc, (MVMObject*)seen_table[log_index]));
             else
                 append(ds, " P6int(type object)");
         } else {
@@ -413,6 +419,26 @@ static void dump_callsite(MVMThreadContext *tc, DumpStr *ds, MVMSpeshGraph *g) {
             MVM_free(argname_utf8);
         }
     }
+    if (g->cs->num_pos)
+        append(ds, "Positional flags: ");
+    for (i = 0; i < g->cs->num_pos; i++) {
+        MVMCallsiteEntry arg_flag = g->cs->arg_flags[i];
+
+        if (i)
+            append(ds, ", ");
+
+        if (arg_flag == MVM_CALLSITE_ARG_OBJ) {
+            append(ds, "obj");
+        } else if (arg_flag == MVM_CALLSITE_ARG_INT) {
+            append(ds, "int");
+        } else if (arg_flag == MVM_CALLSITE_ARG_NUM) {
+            append(ds, "num");
+        } else if (arg_flag == MVM_CALLSITE_ARG_STR) {
+            append(ds, "str");
+        }
+    }
+    if (g->cs->num_pos)
+        append(ds, "\n");
     append(ds, "\n");
 }
 
