@@ -1,5 +1,5 @@
 
-#define MVMNUMUNICODEEXTENTS 104
+#define MVM_NUM_UNICODE_EXTENTS 104
 
 
 struct MVMUnicodeNamedValue {
@@ -8,10 +8,10 @@ struct MVMUnicodeNamedValue {
 };
 #define num_unicode_property_keypairs 3600
 
-MVMint32 MVM_unicode_is_in_block(MVMThreadContext *tc, MVMString *str, MVMint64 pos, MVMString *block);
+MVMint32 MVM_unicode_is_in_block(MVMThreadContext *tc, MVMString *str, MVMint64 pos, MVMString *block_name);
 
-#define MVMCODEPOINTNAMESCOUNT 49764
-#define MVMNUMPROPERTYCODES 93
+#define MVM_CODEPOINT_NAMES_COUNT 49764
+#define MVM_NUM_PROPERTY_CODES 93
 
 
 #define num_unicode_property_value_keypairs 3181
@@ -109,4 +109,4 @@ typedef enum {
     MVM_UNICODE_PROPERTY_XID_CONTINUE = 90,
     MVM_UNICODE_PROPERTY_XID_START = 91,
     MVM_UNICODE_PROPERTY_Z = 92,
-} MVM_unicode_property_codes;
\ No newline at end of file
+} MVM_unicode_property_codes;
