@@ -94,6 +94,10 @@ typedef struct MVMCStruct MVMCStruct;
 typedef struct MVMCStructBody MVMCStructBody;
 typedef struct MVMCStructNameMap MVMCStructNameMap;
 typedef struct MVMCStructREPRData MVMCStructREPRData;
+typedef struct MVMCPPStruct MVMCPPStruct;
+typedef struct MVMCPPStructBody MVMCPPStructBody;
+typedef struct MVMCPPStructNameMap MVMCPPStructNameMap;
+typedef struct MVMCPPStructREPRData MVMCPPStructREPRData;
 typedef struct MVMCUnion MVMCUnion;
 typedef struct MVMCUnionBody MVMCUnionBody;
 typedef struct MVMCUnionNameMap MVMCUnionNameMap;
