@@ -361,7 +361,7 @@ sub allocate_bitfield {
         }
     }
     $first_point->{bitfield_width} = $word_offset+1;
-    $h_sections->{num_property_codes} = "#define MVMNUMPROPERTYCODES $index\n";
+    $h_sections->{num_property_codes} = "#define MVM_NUM_PROPERTY_CODES $index\n";
     $allocated
 }
 
@@ -576,7 +576,7 @@ sub emit_codepoints_and_planes {
         "static const MVMuint16 codepoint_bitfield_indexes[$index] = {\n    ".
             stack_lines(\@bitfield_index_lines, ",", ",\n    ", 0, $wrap_to_columns).
             "\n};";
-    $h_sections->{codepoint_names_count} = "#define MVMCODEPOINTNAMESCOUNT $index";
+    $h_sections->{codepoint_names_count} = "#define MVM_CODEPOINT_NAMES_COUNT $index";
     $extents
 }
 
@@ -686,7 +686,6 @@ sub emit_property_value_lookup {
     my $out = "
 static MVMint32 MVM_unicode_get_property_int(MVMThreadContext *tc, MVMint32 codepoint, MVMint64 property_code) {
     MVMuint32 switch_val = (MVMuint32)property_code;
-    MVMint32 result_val = 0; /* we'll never have negatives, but so */
     MVMuint32 codepoint_row = MVM_codepoint_to_row_index(tc, codepoint);
     MVMuint16 bitfield_row;
 
@@ -801,14 +800,21 @@ static const char* MVM_unicode_get_property_str(MVMThreadContext *tc, MVMint32 c
 }
 
 sub emit_block_lookup {
-    my $hout = "MVMint32 MVM_unicode_is_in_block(MVMThreadContext *tc, MVMString *str, MVMint64 pos, MVMString *block);\n";
-    my $out  = "MVMint32 MVM_unicode_is_in_block(MVMThreadContext *tc, MVMString *str, MVMint64 pos, MVMString *block) {
-    MVMGrapheme32 ord = MVM_string_get_grapheme_at_nocheck(tc, str, pos);
-    MVMuint64 size;
-    char *bname = MVM_string_ascii_encode(tc, block, &size);
+    my $hout = "MVMint32 MVM_unicode_is_in_block(MVMThreadContext *tc, MVMString *str, MVMint64 pos, MVMString *block_name);\n";
+    my $out  = "struct UnicodeBlock {
+    MVMGrapheme32 start;
+    MVMGrapheme32 end;
+
+    char *name;
+    size_t name_len;
+    char *alias;
+    size_t alias_len;
+};
+
+static struct UnicodeBlock unicode_blocks[] = {
 ";
 
-    my $else = '';
+    my @blocks;
     each_line('Blocks', sub {
         $_ = shift;
         my ($from, $to, $block_name) = /^(\w+)..(\w+); (.+)/;
@@ -818,18 +824,47 @@ sub emit_block_lookup {
             my $block_len  = length $block_name;
             my $alias_len  = length $alias_name;
             if ($block_len && $alias_len) {
-                $out .= "
-    $else if (ord >= 0x$from && ord <= 0x$to) {
-        return strncmp(\"$block_name\", bname, $block_len) == 0 || strncmp(\"$alias_name\", bname, $alias_len) == 0;
-    }";
-                $else = 'else';
+                push @blocks, "    { 0x$from, 0x$to, \"$block_name\", $block_len, \"$alias_name\", $alias_len }";
             }
         }
     });
 
-    $out .= "
-    return 0;
+    $out .= join(",\n", @blocks) . "\n";
+
+    $out .= "};
+
+static int block_compare(const void *a, const void *b) {
+    MVMGrapheme32 ord = *((MVMGrapheme32 *) a);
+    struct UnicodeBlock *block = (struct UnicodeBlock *) b;
+
+    if (ord < block->start) {
+        return -1;
+    }
+    else if (ord > block->end) {
+        return 1;
+    }
+    else {
+        return 0;
+    }
+}
+
+MVMint32 MVM_unicode_is_in_block(MVMThreadContext *tc, MVMString *str, MVMint64 pos, MVMString *block_name) {
+    MVMGrapheme32 ord = MVM_string_get_grapheme_at_nocheck(tc, str, pos);
+    MVMuint64 size;
+    char *bname = MVM_string_ascii_encode(tc, block_name, &size);
+    MVMint32 in_block = 0;
+
+    struct UnicodeBlock *block = bsearch(&ord, unicode_blocks, sizeof(unicode_blocks) / sizeof(struct UnicodeBlock), sizeof(struct UnicodeBlock), block_compare);
+
+    if (block) {
+        in_block = strncmp(block->name, bname, block->name_len) == 0 ||
+               strncmp(block->alias, bname, block->alias_len) == 0;
+    }
+    MVM_free(bname);
+
+    return in_block;
 }";
+
     $db_sections->{block_lookup} = $out;
     $h_sections->{block_lookup} = $hout;
 }
@@ -845,7 +880,7 @@ static const MVMint32 codepoint_extents[".($num_extents + 1)."][3] = {\n";
                                      $extent->{fate_type},
                                           ($extent->{fate_really}//0));
     }
-    $h_sections->{MVMNUMUNICODEEXTENTS} = "#define MVMNUMUNICODEEXTENTS $num_extents\n";
+    $h_sections->{MVM_NUM_UNICODE_EXTENTS} = "#define MVM_NUM_UNICODE_EXTENTS $num_extents\n";
     $out .= <<"END";
     {0x10FFFE,0}
 };
@@ -860,18 +895,18 @@ static void generate_codepoints_by_name(MVMThreadContext *tc) {
     MVMint32 codepoint = 0;
     MVMint32 codepoint_table_index = 0;
     MVMUnicodeNameRegistry *entry;
-    for (; extent_index < MVMNUMUNICODEEXTENTS; extent_index++) {
+    for (; extent_index < MVM_NUM_UNICODE_EXTENTS; extent_index++) {
         MVMint32 length;
         codepoint = codepoint_extents[extent_index][0];
         length = codepoint_extents[extent_index + 1][0] - codepoint_extents[extent_index][0];
-        if (codepoint_table_index >= MVMCODEPOINTNAMESCOUNT)
+        if (codepoint_table_index >= MVM_CODEPOINT_NAMES_COUNT)
             continue;
         switch (codepoint_extents[extent_index][1]) {
             case $FATE_NORMAL: {
                 MVMint32 extent_span_index = 0;
                 codepoint_table_index = codepoint_extents[extent_index][2];
                 for (; extent_span_index < length
-                    && codepoint_table_index < MVMCODEPOINTNAMESCOUNT; extent_span_index++) {
+                    && codepoint_table_index < MVM_CODEPOINT_NAMES_COUNT; extent_span_index++) {
                     const char *name = codepoint_names[codepoint_table_index];
                     if (name) {
                         MVMUnicodeNameRegistry *entry = MVM_malloc(sizeof(MVMUnicodeNameRegistry));
